from typing import Union
from .nfa import NFA
from .dfa import DFA


def automaton_to_graphviz(automaton: Union[NFA, DFA], path: str) -> None:
    """
    Converts automaton to the graphviz format (.dot file) and saves
    the result in the specified location.

    Args:
        automaton (Union[NFA, DFA]): Automaton to be converted.
        path (str): Path where the file will be created.
    """
    with open(path, "w", encoding="utf-8") as file:
        file.write("digraph G {\n")
        file.write('\trankdir="LR"\n')
        file.write('\t__init__[shape=none label=""]\n')
        file.write(f"\t__init__ -> {automaton.initial_state}\n")

        for s_from in automaton.states:
            for s_to in automaton.states:
                label_symbols = automaton.get_symbols_between_states(s_from, s_to)
                if "" in label_symbols:
                    label_symbols.discard("")
                    label_symbols.add("ε")

                if label := ",".join(label_symbols):
                    file.write(f'\t{s_from} -> {s_to}[label="{label}"]\n')

        for non_fin_state in automaton.states.difference(automaton.final_states):
            file.write(f"\t{non_fin_state} [shape=circle]\n")

        for fin_state in automaton.final_states:
            file.write(f"\t{fin_state} [shape=doublecircle]\n")

        file.write("}\n")


if __name__ == "__main__":
    pass



from enum import Enum
from typing import List

START_SYMBOL = ">"
EMPTY_SYMBOL = ""


class Direction(Enum):
    LEFT = -1
    STAY = 0
    RIGHT = 1

    def __repr__(self):
        return self.name


class Cell:
    """Represents one cell of a Turing machine memory tape."""

    def __init__(
        self, value: str = EMPTY_SYMBOL, right: "Cell" = None, left: "Cell" = None
    ) -> None:
        self.value = value
        self.left = left
        self.right = right

    def __repr__(self):
        return self.value or " "


class Tape:
    """Represents Turing machine memory tape as a linked list."""

    def __init__(self, start: Cell = None):
        self.start = start or Cell()
        self.current = self.start

    def __repr__(self) -> str:
        curr_cell: Cell = self.start

        # find the leftmost cell
        while curr_cell.left:
            curr_cell = curr_cell.left

        cells: List[Cell] = []

        while curr_cell:
            cells.append(curr_cell)
            curr_cell = curr_cell.right

        # ['a', 'b'] -> | a | b |
        str_cells = f"| {' | '.join([repr(c) for c in cells])} |\n"

        # adds spaces to put the cursor '^' below the current cell
        str_cursor = f"{' ' * (2 + cells.index(self.current) * 4)}^\n"

        return str_cells + str_cursor

    def move(self, direction: Direction) -> None:
        """Moves the tape cursor based on the provided direction.

        Args:
            direction (Direction): Specifies the move direction.
        """
        if direction == Direction.LEFT:
            self.move_left()
        elif direction == Direction.RIGHT:
            self.move_right()

    def move_left(self):
        """Moves the tape cursor to the left. (current will be current.left)"""
        if not self.current.left:
            self.current.left = Cell(right=self.current)

        self.current = self.current.left

    def move_right(self):
        """Moves the tape cursor to the right. (current will be current.right)"""
        if not self.current.right:
            self.current.right = Cell(left=self.current)

        self.current = self.current.right

    def write(self, text: str) -> None:
        """Writes the provided text on the tape.

        Args:
            text (str): Text to be written on the tape.
        """
        for symbol in text:
            self.current.value = symbol
            self.move_right()

        self.current = self.start

    def write_symbol(self, symbol: str) -> None:
        """Overwrites the current cell with the provided symbol.

        Args:
            symbol (str): Symbol to be written on the current cell.
        """
        cell = self.current

        if cell.value == START_SYMBOL:
            return

        self.current.value = symbol

    def clear(self) -> None:
        """Clears the tape contents and places the cursor on the start."""
        self.start = Cell()
        self.current = self.start


if __name__ == "__main__":
    pass



from typing import Set


class BaseTuringMachine:
    """Represents an abstract Turing machine class. This class cannot be instantiated."""

    def __new__(cls, *args, **kwargs):
        if cls is BaseTuringMachine:
            raise TypeError("Only DTM and MTM can be instantiated!")

        return object.__new__(cls)

    def __init__(
        self,
        states: Set[str],
        acc_states: Set[str],
        rej_states: Set[str] = set(),
        initial_state: str = "init",
        start_symbol: str = ">",
        empty_symbol: str = "",
    ) -> None:
        self.states = states
        self.acc_states = acc_states
        self.rej_states = rej_states
        self.initial_state = initial_state
        self.start_symbol = start_symbol
        self.empty_symbol = empty_symbol

        # After exceeding max_steps value, the turing machine will be considered as looping.
        # Change this value for more complex scenarios.
        self.max_steps = 100

    def add_state(self, state: str, is_acc: bool = False, is_rej: bool = False) -> bool:
        """Adds state to the machine.

        Args:
            state (str): State to be added.
            is_acc (bool, optional): True if the state is accepting. Defaults to False.
            is_rej (bool, optional): True if the state is rejecting. Defaults to False.

        Returns:
            bool: False if the state is already present or 'is_acc' and 'is_rej' arguments are both True.
                  True otherwise.
        """
        if state in self.states or (is_acc and is_rej):
            return False

        if is_acc:
            self.acc_states.add(state)

        if is_rej:
            self.rej_states.add(state)

        return True

    def remove_state(self, state: str) -> bool:
        """
        Removes the state from the machine.

        Args:
            state (str): State to be removed.

        Returns:
            bool: Returns True if the state was added successfully, False otherwise.
        """
        if state not in self.states:
            return False

        if state in self.acc_states:
            self.acc_states.remove(state)

        if state in self.rej_states:
            self.rej_states.remove(state)

        self.states.remove(state)

        return True

    def complement(self) -> None:
        """
        Complements the automaton. (Final states will become non-final and vice-versa).
        """
        self.final_states = self.states - self.final_states


if __name__ == "__main__":
    pass



from os import name, system
from time import sleep
from typing import Dict, Set, Tuple

from .base import BaseTuringMachine
from .tape import Direction, Tape

DTMRule = Tuple[str, str, Direction]
DTMRules = Dict[str, DTMRule]
DTMTransitions = Dict[str, DTMRules]


class DTM(BaseTuringMachine):
    """Represents a DETERMINISTIC Turing Machine"""

    def __init__(
        self,
        states: Set[str],
        input_alphabet: Set[str],
        acc_states: Set[str],
        rej_states: Set[str] = set(),
        transition_function: DTMTransitions = None,
        tape: Tape = Tape(),
        initial_state: str = "init",
        start_symbol: str = ">",
        empty_symbol: str = "",
    ) -> None:
        if transition_function is None:
            transition_function = {}
        super().__init__(
            states,
            acc_states,
            rej_states,
            initial_state,
            start_symbol,
            empty_symbol,
        )
        self.input_alphabet = input_alphabet
        self.transition_function = transition_function
        self.tape = tape

    def get_transition(self, state: str, read: str) -> DTMRule:
        return self.transition_function.get(state, {}).get(read, None)

    def remove_state(self, state: str) -> bool:
        if not super().remove_state(state):
            return False

        for state in self.transition_function:
            if state in self.transition_function.keys():
                del self.transition_function[state]

            for read in self.transition_function[state]:
                next_s, _, _ = self.transition_function[state][read]
                if next_s == state:
                    del self.transition_function[state][read]

        return True

    def simulate(
        self,
        to_console: bool = True,
        to_file: bool = False,
        path: str = "simulation.txt",
        delay: float = 0.5,
    ) -> bool:
        """Simulates the machine on its current tape configuration.

        Args:
            to_console (bool, optional): Set to False if you only want to see the result. Defaults to True.
            to_file (bool, optional): Set to True if you want to save every step to the file. Defaults to False.
            path (str, optional): Path to the file with the step history. Defaults to "simulation.txt".
            delay (float, optional): The delay (s) between each step when printing to console. Defaults to 0.5.

        Returns:
            bool: False if the machine rejects the word or exceeds the 'max_steps' value, True otherwise.
        """
        state: str = self.initial_state
        steps: int = 1
        output_file = None
        step_separator = f"\n{'=' * 40}\n\n"

        def clear_console() -> None:
            if name == "posix":
                system("clear")
            else:
                system("cls")

        def get_rule_string() -> str:
            row = self.get_transition(state, self.tape.current.value)
            step_index = steps if row else steps - 1
            next_step = f"{state}, {self.tape.current.value}"

            return f"{step_index}. ({next_step}) -> {row}\n"

        def print_automaton_state() -> None:
            rule_str = get_rule_string()

            if output_file:
                output_file.write(rule_str)
                output_file.write(repr(self.tape))
                output_file.write(step_separator)

            if to_console:
                clear_console()
                print(rule_str, "\n", self.tape)

                sleep(delay)

        def close():
            print_automaton_state()

            if output_file:
                output_file.close()

        if to_file:
            output_file = open(path, "w")

        while steps <= self.max_steps:
            print_automaton_state()
            if state in self.acc_states:
                close()
                return True

            rule = self.get_transition(state, self.tape.current.value)

            if not rule or rule[0] in self.rej_states:
                close()
                return False

            steps += 1
            state, write, direction = rule
            self.tape.write_symbol(write)
            self.tape.move(direction)

        close()

        print(f"Exceeded the maximum allowed steps. ({self.max_steps})")
        print(
            "You change the default value by setting the 'max_steps' property of this automaton."
        )

        return False


if __name__ == "__main__":
    pass



from typing import Dict, Set

from .base import BaseFiniteAutomaton

NFARules = Dict[str, Set[str]]
NFATransitions = Dict[str, NFARules]


class NFA(BaseFiniteAutomaton):
    """
    Nondeterministic Finite Automaton
    """

    def __init__(
        self,
        states: Set[str],
        alphabet: Set[str],
        initial_state: str,
        final_states: Set[str],
        transitions: NFATransitions,
    ):
        super().__init__(states, alphabet, initial_state, final_states)

        self.transitions = transitions

    def __repr__(self) -> str:
        return super().__repr__() + "\n" + self.__repr_transitions__()

    def __repr_transitions__(self) -> str:
        def get_max_cell_width():
            max_cell_width = 5

            for key, value in self.transitions.items():
                for next_value in value.values():
                    width = len(",".join(sorted(next_value)))
                    max_cell_width = max(max_cell_width, width, len(key))

            # add 4 for spaces on both sides
            return max_cell_width + 4

        cell_width = get_max_cell_width()
        header = f"{'NFA': ^{cell_width+5}}|"
        for letter in sorted(self.alphabet):
            header += f"{letter or 'ε': ^{cell_width}}|"

        rows = ""
        for state in sorted(self.states):
            row_prefix = self.__repr_row_prefix__(state)
            rows += f"{row_prefix: ^5}{state or 'empty': <{cell_width}}"

            for letter in sorted(self.alphabet):
                transition = self.get_transition(state, letter)
                transition_str = "{" + ",".join(transition) + "}"
                rows += f"|{transition_str or 'empty': ^{cell_width}}"

            rows += "\n"

        divider = "-" * rows.index("\n")

        return "\n".join([header[:-1], divider, rows])

    def get_transition(self, state_from: str, symbol: str) -> Set[str]:
        """
        Returns next possible states from the provided state by symbol.

        Args:
            state_from (str): State name where the transition starts.
            symbol (str): Transition symbol.

        Returns:
            Set[str]: Set of next states.
        """
        return self.transitions.get(state_from, {}).get(symbol, set())

    def set_transition(self, state_from: str, states_to: Set[str], symbol: str) -> None:
        """
        Adds/changes transition to automaton. And returns bool value based on success.
        If the automaton already contains transition from 'state_from' by 'symbol', it will be overwritten.

        Args:
            state_from (str): State name where the transition starts.
            states_to (Set[str]): States where the transition ends.
            symbol (str): Transition symbol.

        Returns:
            bool: True if transition was added, False otherwise.
        """
        if state_from not in self.transitions.keys():
            self.transitions[state_from] = {symbol: states_to}
        else:
            self.transitions[state_from][symbol] = states_to

    def add_transition(self, state_from: str, states_to: Set[str], symbol: str) -> bool:
        """
        Adds transition to the automaton and returns bool based on a change.

        Args:
            state_from (str): State name where the transition starts.
            states_to (Set[str]): Set of states where the transition ends.
            symbol (str): Transition symbol.

        Returns:
            bool: True if the transition function changed, otherwise False.
        """
        if state_from not in self.transitions.keys():
            self.transitions[state_from] = {symbol: states_to}
            return True

        transition = self.get_transition(state_from, symbol)

        if not transition:
            self.transitions[state_from][symbol] = states_to
            return len(states_to) > 0

        if not states_to - transition:
            return False

        self.transitions[state_from][symbol].update(states_to)

        return True

    def remove_transition(self, state_from: str, state_to: str, symbol: str) -> bool:
        """
        Adds transition to the automaton and returns bool based on success.

        Args:
            state_from (str): State name where the transition starts.
            state_to (str): State name where the transition ends.
            symbol (str): Transition symbol.

        Returns:
            bool: True if transition was successfully added, False otherwise
        """
        transition = self.get_transition(state_from, symbol)

        if not transition or state_to not in transition:
            return False

        self.transitions[state_from][symbol].difference_update({state_to})

        if not self.transitions[state_from][symbol]:
            del self.transitions[state_from][symbol]

        return True

    def get_symbols_between_states(self, state_from: str, state_to: str) -> Set[str]:
        """
        Returns set of symbols between two states.

        Args:
            state_from: State name where the transition starts.
            state_to: State name where the transition ends.

        Returns:
            Set of symbols.
        """
        return {
            s
            for s in self.transitions.get(state_from, {}).keys()
            if state_to in self.get_transition(state_from, s)
        }

    def add_state(self, state: str, is_final: bool = False) -> bool:
        """
        Adds a state to the automaton.

        Args:
            state (str): Name of the state to be added.
            is_final (bool, optional): Whether to mark the state as final. Defaults to False.

        Returns:
            bool: True if automaton did not contain such state, False otherwise.
        """
        return super().add_state(state, is_final)

    def remove_state(self, state: str) -> bool:
        """
        Removes provided state from the automaton (from its states and transitions)

        Args:
            state (str): state to be removed
        """
        if not super().remove_state(state):
            return False
        
        self.transitions.pop(state, None)

        for s in self.transitions.keys():
            rules = self.transitions[s]

            for k in rules.keys():
                rules[k] = rules[k] - {state}

        return True

    def is_accepted(self, input_string: str) -> bool:
        """
        Checks whether the provided input string is accepted by the automaton.

        Args:
            input_string (str): Input string to be tested.

        Returns:
            bool: True if word is accepted, False otherwise.
        """

        def is_accepted_rec(current_state: str, curr_input: str) -> bool:
            if not curr_input:
                return current_state in self.final_states

            if not self.get_transition(current_state, curr_input[0]):
                return False

            result = False

            for state in self.get_transition(current_state, curr_input[0]):
                result = result or is_accepted_rec(state, curr_input[1:])

            return result

        return is_accepted_rec(self.initial_state, input_string)

    def is_valid(self) -> bool:
        """
        Checks whether the NFA is valid:
            1. States set is not empty.
            2. Initial state is in states.
            3. Final states are subset of states.
            4. The transition function contains characters only from its alphabet.
            5. The transition function contains states only from its states set.

        Returns:
            bool: True if NFA is valid, False otherwise.
        """
        # creates set of states used in the transition function
        used_states = set().union(
            *sum((list(v.values()) for v in list(self.transitions.values())), [])
        ) | set(self.transitions.keys())

        # rule 5
        if used_states - self.states:
            return False

        # rules 1-4
        if not super().is_valid() or any(
            set(rule.keys()) - self.alphabet for rule in self.transitions.values()
        ):
            return False

        return True


if __name__ == "__main__":
    pass
